# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Help Plugin API.
"""

# Local imports
from spyder.plugins.help.plugin import HelpActions
from spyder.plugins.help.widgets import (HelpWidgetActions,
                                         HelpWidgetMainToolbarSections,
                                         HelpWidgetOptionsMenuSections)
