# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Local imports
from spyder.plugins.onlinehelp.widgets import PydocBrowserActions
from spyder.widgets.browser import WebViewActions
